#include "FM10K.h"
#include "registers/Register.h"
#include "IES.h"
#include "device/PCIEDevice.h"
#include <chrono>
#include <thread>
#include <iostream>
#include <netdb.h>
#include "ryml.hpp"
#include "c4/std/string.hpp"
#include "IES_SDK.h"


FM10K::FM10K::FM10K(std::unique_ptr<Device> device) : m_device(std::move(device)) {
    if(valid()){
        auto vpd = mapRegister<registers::MGMT::VITAL_PRODUCT_DATA>();
        auto chipVersion = mapRegister<registers::MGMT::CHIP_VERSION>();
        auto fuseData = mapRegister<registers::MGMT::FUSE_DATA_0>();

        m_hwInfo = {
                static_cast<HardwareInformation::Family>(vpd->fields.PartNumber),
                static_cast<HardwareInformation::Model>(fuseData->fields.SKU),
                static_cast<HardwareInformation::ChipVersion>(chipVersion->fields.Version),
                };
    }
}


const FM10K::FM10K::HardwareInformation& FM10K::FM10K::getHardwareInformation() const {
    return m_hwInfo;
}



std::string FM10K::FM10K::getHardwareInformationString() const {
    auto& info = getHardwareInformation();
    std::string s;
    s.reserve(2);

    if(info.family != HardwareInformation::Family::FM10000){
        return "unknown";
    }

    switch (info.model) {
        default:
            s += "FM10000-unknown";
            break;
        case HardwareInformation::Model::FM10840:
            s += "FM10840";
            break;
        case HardwareInformation::Model::FM10420:
            s += "FM10420";
            break;
        case HardwareInformation::Model::FM10824:
            s += "FM10824";
            break;
        case HardwareInformation::Model::FM10064:
            s += "FM10064";
            break;
        case HardwareInformation::Model::FM10036:
            s += "FM10036";
            break;
    }

    switch (info.chipVersion) {
        default:
            s += "_unknown";
            break;
        case HardwareInformation::ChipVersion::A0:
            s += "_A0";
            break;
        case HardwareInformation::ChipVersion::B0:
            s += "_B0";
            break;
    }

    return s;
}

bool FM10K::FM10K::isPEPEnabled(uint8_t pep) const {
    return (mapRegister<registers::MGMT::DEVICE_CFG>()->fields.PCIeEnable & (1 << pep)) > 0;
}

void FM10K::FM10K::wait(uint32_t microseconds) {
    std::this_thread::sleep_for(std::chrono::microseconds(microseconds));
}

void FM10K::FM10K::wait_ms(uint32_t milliseconds) {
    std::this_thread::sleep_for(std::chrono::milliseconds (milliseconds));
}

void FM10K::FM10K::wait_ns(uint32_t nanoseconds) {
    std::this_thread::sleep_for(std::chrono::nanoseconds (nanoseconds));
}

std::string getYAMLString(c4::yml::NodeRef ref, const std::string& def = ""){
    if(ref.valid() && !ref.is_seed() && ref.has_val()){
        std::string s;
        c4::from_chars(ref.val(), &s);
        return s;
    }
    return def;
}

int32_t getYAMLInteger(c4::yml::NodeRef ref, int32_t def = 0){
    if(ref.valid() && !ref.is_seed() && ref.has_val() && ref.val().is_integer()){
        int32_t s;
        if(atoi_first(ref.val(), &s)){
            return s;
        }
    }
    return def;
}

int64_t getYAMLInteger64(c4::yml::NodeRef ref, int64_t def = 0){
    if(ref.valid() && !ref.is_seed() && ref.has_val() && ref.val().is_integer()){
        int64_t s;
        if(atoi_first(ref.val(), &s)){
            return s;
        }
    }
    return def;
}

uint32_t getYAMLUnsignedInteger(c4::yml::NodeRef ref, uint32_t def = 0){
    if(ref.valid() && !ref.is_seed() && ref.has_val() && ref.val().is_integer()){
        uint32_t s;
        if(atou_first(ref.val(), &s)){
            return s;
        }
    }
    return def;
}

uint64_t getYAMLUnsignedInteger64(c4::yml::NodeRef ref, uint64_t def = 0){
    if(ref.valid() && !ref.is_seed() && ref.has_val() && ref.val().is_integer()){
        uint64_t s;
        if(atou_first(ref.val(), &s)){
            return s;
        }
    }
    return def;
}

bool getYAMLBool(c4::yml::NodeRef ref, bool def = false){
    if(ref.valid() && !ref.is_seed() && ref.has_val()){
        if(ref.val() == "true"){
            return true;
        } else if(ref.val() == "false"){
            return false;
        }
    }
    return def;
}

static std::unordered_map<std::string, fm_int> STP_STATE_MAP = {
        {"disabled", FM_STP_STATE_DISABLED},
        {"listening", FM_STP_STATE_LISTENING},
        {"learning", FM_STP_STATE_LEARNING},
        {"forwarding", FM_STP_STATE_FORWARDING},
        {"blocking", FM_STP_STATE_BLOCKING},
};

static std::unordered_map<std::string, fm_int> TAGGING_MODE_MAP = {
        {"8021q", FM_PORT_TAGGING_8021Q},
        {"8021ad_customer", FM_PORT_TAGGING_8021AD_CUST},
        {"8021ad_provider", FM_PORT_TAGGING_8021AD_PROV},
        {"pseudo1", FM_PORT_TAGGING_PSEUDO1},
        {"pseudo2", FM_PORT_TAGGING_PSEUDO2},
};

static std::unordered_map<std::string, fm_int> ISL_FORMAT_MAP = {
        {"none", FM_ISL_TAG_NONE},
        {"f32", FM_ISL_TAG_F32},
        {"f56", FM_ISL_TAG_F56},
        {"f64", FM_ISL_TAG_F64},
        {"f96", FM_ISL_TAG_F96},
        {"other_32", FM_ISL_TAG_OTHER_32B},
        {"other_64", FM_ISL_TAG_OTHER_64B},
        {"other_96", FM_ISL_TAG_OTHER_96B},
};

static std::unordered_map<std::string, fm_int> LOOPBACK_MODE_MAP = {
        {"off", FM_PORT_LOOPBACK_OFF},
        {"rx2tx", FM_PORT_LOOPBACK_RX2TX},
        {"tx2rx", FM_PORT_LOOPBACK_TX2RX},
};


static std::unordered_map<std::string, fm_int> PARSER_MODE_MAP = {
        {"l2", FM_PORT_PARSER_STOP_AFTER_L2},
        {"l3", FM_PORT_PARSER_STOP_AFTER_L3},
        {"l4", FM_PORT_PARSER_STOP_AFTER_L4},
        {"all", FM_PORT_PARSER_PARSE_ALL},
};


static std::unordered_map<std::string, fm_int> BFLOOD_MODE_MAP = {
        {"forward_without_cpu", FM_PORT_BCAST_FWD_EXCPU},
        {"forward", FM_PORT_BCAST_FWD},
        {"discard", FM_PORT_BCAST_DISCARD},
        {"trap", FM_PORT_BCAST_TRAP},
};


static std::unordered_map<std::string, fm_int> MFLOOD_MODE_MAP = {
        {"forward_without_cpu", FM_PORT_MCAST_FWD_EXCPU},
        {"forward", FM_PORT_MCAST_FWD},
        {"discard", FM_PORT_MCAST_DISCARD},
        {"trap", FM_PORT_MCAST_TRAP},
};


static std::unordered_map<std::string, fm_int> UFLOOD_MODE_MAP = {
        {"forward_without_cpu", FM_PORT_UCAST_FWD_EXCPU},
        {"forward", FM_PORT_UCAST_FWD},
        {"discard", FM_PORT_UCAST_DISCARD},
        {"trap", FM_PORT_UCAST_TRAP},
};


static std::unordered_map<std::string, fm_int> VLAN_LEARNING_MODE = {
        {"independent", FM_VLAN_LEARNING_MODE_INDEPENDENT},
        {"shared", FM_VLAN_LEARNING_MODE_SHARED}
};


static std::unordered_map<std::string, fm_int> STP_MODE = {
        {"shared", FM_SPANNING_TREE_SHARED},
        {"per_vlan", FM_SPANNING_TREE_PER_VLAN},
        {"multiple", FM_SPANNING_TREE_MULTIPLE},
};


static std::unordered_map<std::string, fm_int> LAG_MODE = {
        {"static", FM_MODE_STATIC},
        {"dynamic", FM_MODE_DYNAMIC},
};


static std::unordered_map<std::string, fm_int> AUTONEG_MAP = {
        {"none", FM_PORT_AUTONEG_NONE},
        {"sgmii", FM_PORT_AUTONEG_SGMII},
        {"clause37", FM_PORT_AUTONEG_CLAUSE_37},
        {"clause73", FM_PORT_AUTONEG_CLAUSE_73},
};


static std::unordered_map<std::string, fm_int> ETHMODE_MAP = {
        {"10G-SR", FM_ETH_MODE_10GBASE_SR},
        {"25G-SR", FM_ETH_MODE_25GBASE_SR},
        {"40G-SR4", FM_ETH_MODE_40GBASE_SR4},
        {"100G-SR4", FM_ETH_MODE_100GBASE_SR4},
};


static std::unordered_map<std::string, fm_int> ROUTE_ENTRY_TYPE_MAP = {
        {"unicast", FM_ROUTE_TYPE_UNICAST},
        {"unicast_ecmp", FM_ROUTE_TYPE_UNICAST_ECMP},
        {"multicast", FM_ROUTE_TYPE_MULTICAST},
};


static std::unordered_map<std::string, fm_int> ROUTER_MAC_MODE_MAP = {
        {"physical", FM_ROUTER_MAC_MODE_PHYSICAL_MAC_ADDRESS},
        {"virtual1", FM_ROUTER_MAC_MODE_VIRTUAL_MAC_ADDRESS_1},
        {"virtual2", FM_ROUTER_MAC_MODE_VIRTUAL_MAC_ADDRESS_1},
        {"virtual12", FM_ROUTER_MAC_MODE_VIRTUAL_MAC_ADDRESS_1_AND_2},
};


static std::unordered_map<std::string, fm_int> ACL_FRAME_TYPE_MAP = {
        {"non_ip", FM_ACL_FRAME_TYPE_NON_IP},
        {"ipv4", FM_ACL_FRAME_TYPE_IPV4},
        {"ipv6", FM_ACL_FRAME_TYPE_IPV6},
};

fm_int getMapState(const std::string& k, const std::unordered_map<std::string, fm_int>& m){
    assert(m.find(k) != m.end());
    return m.at(k);
}

fm_ipAddr stringToIpAddr(const char* ip){
    struct addrinfo* addr;

    if(getaddrinfo(ip, nullptr, nullptr, &addr) != 0){
        throw std::runtime_error("Invalid address");
    }

    if(addr->ai_next != nullptr || (addr->ai_family != AF_INET && addr->ai_family != AF_INET6)){
        freeaddrinfo(addr);
        throw std::runtime_error("Invalid address type");
    }

    fm_ipAddr result;

    result.isIPv6 = addr->ai_family == AF_INET6;
    if(result.isIPv6){
        auto ip6 = reinterpret_cast<struct sockaddr_in6*>(addr->ai_addr);
        result.addr[0] = ip6->sin6_addr.s6_addr32[0];
        result.addr[1] = ip6->sin6_addr.s6_addr32[1];
        result.addr[2] = ip6->sin6_addr.s6_addr32[2];
        result.addr[3] = ip6->sin6_addr.s6_addr32[3];
    }else{
        auto ip4 = reinterpret_cast<struct sockaddr_in*>(addr->ai_addr);
        result.addr[0] = ip4->sin_addr.s_addr;
    }

    freeaddrinfo(addr);

    return result;
}

std::vector<uint8_t> HexToBytes(const std::string& hex) {
    std::vector<uint8_t> bytes;

    for (unsigned int i = 0; i < hex.length(); i += 2) {
        std::string byteString = hex.substr(i, 2);
        char byte = (char) strtol(byteString.c_str(), NULL, 16);
        bytes.push_back(byte);
    }

    return std::move(bytes);
}

template <typename T>
void setACLMatchValueMask(const ryml::NodeRef& node, T& value, T& mask){
    if(node.is_seq()){
        if constexpr(std::is_same_v<T, fm_ipAddr>){
            value = static_cast<T>(stringToIpAddr(getYAMLString(node[0]).c_str()));
            mask = static_cast<T>(stringToIpAddr(getYAMLString(node[1]).c_str()));
        }else if constexpr(std::is_array<T>::value && std::is_same_v<std::remove_extent<T>, uint8_t>) {
            auto data = HexToBytes(getYAMLString(node[0]));
            auto mask_v = HexToBytes(getYAMLString(node[1]));
            assert((data.size() / sizeof(std::remove_extent<T>)) < sizeof(T));
            assert((mask_v.size() / sizeof(std::remove_extent<T>)) < sizeof(T));

            for(auto i = 0; i < sizeof(T); ++i){
                value[i] = i < data.size() ? data[i] : 0;
                mask[i] = i < mask_v.size() ? mask_v[i] : 0;
            }
        }else if constexpr( sizeof(T) == 1){
            value = static_cast<T>(getYAMLUnsignedInteger(node[0]) & 0xFF);
            mask = static_cast<T>(getYAMLUnsignedInteger(node[1]) & 0xFF);
        }else if constexpr( sizeof(T) == 2){
            value = static_cast<T>(getYAMLUnsignedInteger(node[0]) & 0xFFFF);
            mask = static_cast<T>(getYAMLUnsignedInteger(node[1]) & 0xFFFF);
        }else if constexpr( sizeof(T) == 4){
            value = static_cast<T>(getYAMLUnsignedInteger(node[0]));
            mask = static_cast<T>(getYAMLUnsignedInteger(node[1]));
        }else if constexpr (sizeof(T) == 8){
            value = static_cast<T>(getYAMLUnsignedInteger64(node[0]));
            mask = static_cast<T>(getYAMLUnsignedInteger64(node[1]));
        }else{
            std::abort();
        }
    }else{
        //No mask set

        if constexpr(std::is_same_v<T, fm_ipAddr>){
            value = static_cast<T>(stringToIpAddr(getYAMLString(node).c_str()));
            auto ptr = reinterpret_cast<unsigned char*>(&mask.addr);
            std::fill(ptr, ptr + sizeof(mask), 0xFF);
            mask.isIPv6 = value.isIPv6;
        }else if constexpr(std::is_array<T>::value && std::is_same_v<std::remove_extent<T>, uint8_t>) {
            auto data = HexToBytes(getYAMLString(node));
            assert((data.size() / sizeof(std::remove_extent<T>)) < sizeof(T));

            for(auto i = 0; i < sizeof(T); ++i){
                value[i] = i < data.size() ? data[i] : 0;
                mask[i] = 0xFF;
            }
        }else if constexpr( sizeof(T) == 1){
            auto ptr = reinterpret_cast<unsigned char*>(&mask);
            std::fill(ptr, ptr + sizeof(mask), 0xFF);
            value = static_cast<T>(getYAMLUnsignedInteger(node) & 0xFF);
        }else if constexpr( sizeof(T) == 2){
            auto ptr = reinterpret_cast<unsigned char*>(&mask);
            std::fill(ptr, ptr + sizeof(mask), 0xFF);
            value = static_cast<T>(getYAMLUnsignedInteger(node) & 0xFFFF);
        }else if constexpr( sizeof(T) == 4){
            auto ptr = reinterpret_cast<unsigned char*>(&mask);
            std::fill(ptr, ptr + sizeof(mask), 0xFF);
            value = static_cast<T>(getYAMLUnsignedInteger(node));
        }else if constexpr (sizeof(T) == 8){
            auto ptr = reinterpret_cast<unsigned char*>(&mask);
            std::fill(ptr, ptr + sizeof(mask), 0xFF);
            value = static_cast<T>(getYAMLUnsignedInteger64(node));
        }else{
            std::abort();
        }
    }
}

bool FM10K::FM10K::initialize(const ryml::Tree& tree) const {
    auto& ies = IES::getInstance();

    auto root = tree.rootref();

    if(!root["platform"].is_seed() && root["platform"].is_map()){
        if(!root["platform"]["configuration"]["preset"].is_seed()){
            auto conf = getYAMLString(root["platform"]["configuration"]["preset"]);
            std::cout << "Loading know configuration " << conf << std::endl;
            if(!ies.loadKnownConfiguration(conf)){
                return false;
            }
        }
        if(!root["platform"]["configuration"]["entries"].is_seed() && root["platform"]["configuration"]["entries"].is_seq()){
            for(auto e : root["platform"]["configuration"]["entries"]){
                if(e.is_seq() && e.num_children() == 3){ //Conf mapping
                    auto key = getYAMLString(e[0]);
                    auto type = e[1];
                    auto value = getYAMLString(e[2]);
                    if(type == "text"){
                        ies.replaceConfigEntry(key, IES::TlvType::Text, value);
                    }else if(type == "int"){
                        ies.replaceConfigEntry(key, IES::TlvType::Integer, value);
                    }else if(type == "bool"){
                        ies.replaceConfigEntry(key, IES::TlvType::Boolean, value);
                    }else if(type == "null"){
                        ies.replaceConfigEntry(key, IES::TlvType::Null, value);
                    }else{
                        std::abort();
                    }
                }
            }
        }
    }

    auto pcieDevice = dynamic_cast<PCIEDevice*>(m_device.get());
    if(pcieDevice != nullptr){
        ies.replaceConfigEntry("api.platform.config.switch.0.uioDevName", IES::TlvType::Null, "");
        ies.replaceConfigEntry("api.platform.config.switch.0.resource4DevName", IES::TlvType::Text, pcieDevice->getDeviceEntry().getPath());
    }else{
        return false;
    }
    //TODO UIO support


    //ies.addConfigEntry("api.platform.lib.config.debug", IES::TlvType::Text, "I2C_RW,I2C_MUX,DUMP_CFG");
    //ies.addConfigEntry("api.debug.initLoggingCategories", IES::TlvType::Text, "FM_LOG_CAT_LOGGING|FM_LOG_CAT_LINK_STATE|FM_LOG_CAT_ALOS|FM_LOG_CAT_DEBUG|FM_LOG_CAT_PHY|FM_LOG_CAT_PLATFORM|FM_LOG_CAT_EVENT|FM_LOG_CAT_SWITCH|FM_LOG_CAT_MAILBOX|FM_LOG_CAT_PORT|FM_LOG_CAT_SERDES|FM_LOG_CAT_EVENT_INTR|FM_LOG_CAT_STATE_MACHINE|FM_LOG_CAT_GENERAL|FM_LOG_CAT_API|FM_LOG_CAT_PORT_AUTONEG");

    if(ies.init()){
        if(!root["vlans"].is_seed() && root["vlans"].is_seq()){
            for(auto vlan : root["vlans"]){
                if(!vlan.is_map()){
                    std::abort();
                }
                auto vid = getYAMLUnsignedInteger(vlan["id"], 0xFFFF);
                if(vid == 0xFFFF){
                    std::abort();
                }

                std::cout << "Creating VLAN " << vid;
                ies.fmCreateVlan(vid);

                if(!vlan["attributes"].is_seed() && vlan["attributes"].is_map()){
                    for(auto attr : vlan["attributes"]){
                        std::string key;
                        c4::from_chars(attr.key(), &key);

                        if(key == "reflect"){
                            auto val = getYAMLBool(attr, true);
                            std::cout << " " << key << " = " << (val ? "true" : "false");
                            ies.fmSetVlanAttribute(vid, FM_VLAN_REFLECT, val);
                        }else if(key == "routable"){
                            auto val = getYAMLBool(attr, true);
                            std::cout << " " << key << " = " << (val ? "true" : "false");
                            ies.fmSetVlanAttribute(vid, FM_VLAN_ROUTABLE, val);
                        }
                    }
                }

                std::cout << std::endl;
            }
        }

        if(!root["ports"].is_seed() && root["ports"].is_seq()){

            auto ports = root["ports"];



            auto applyPortSettings = [&ies](c4::yml::NodeRef portEntry, int logicalPort, int physicalPort, int cpi, bool isInternalPort, bool isPCIePort, bool isSpecialPort, bool isEplPort, bool isCpuPort){

                if(!portEntry["internal"].is_seed()){
                    ies.fmSetPortAttribute(logicalPort, FM_PORT_INTERNAL, getYAMLBool(portEntry["internal"], false));
                }
                if(!portEntry["vlans"].is_seed() && portEntry["vlans"].is_seq()){
                    for (auto portVlan : portEntry["vlans"]) {
                        auto vid = getYAMLUnsignedInteger(portVlan["id"], 0xFFFF);
                        if(vid == 0xFFFF){
                            std::abort();
                        }

                        std::cout << "Add port " << logicalPort << " to vlan " << vid << std::endl;
                        ies.fmAddVlanPort(vid, logicalPort, !isPCIePort && getYAMLBool(portVlan["tag"], true));
                        ies.fmSetVlanPortState(vid, logicalPort, getMapState(getYAMLString(portVlan["stp"], "forwarding"), STP_STATE_MAP));
                    }
                }
                if(!portEntry["attributes"].is_seed() && portEntry["attributes"].is_map()){
                    for (auto attr : portEntry["attributes"]) {
                        std::string key;
                        c4::from_chars(attr.key(), &key);

                        if(key == "learning"){
                            ies.fmSetPortAttribute(logicalPort, FM_PORT_LEARNING, getYAMLBool(attr, !(isPCIePort || isSpecialPort)));
                        }else if(key == "learning"){
                            ies.fmSetPortAttribute(logicalPort, FM_PORT_MAX_FRAME_SIZE, getYAMLUnsignedInteger(attr, FM10000_PCIE_MAX_FRAME_SIZE));
                        }else if(key == "switch_priority"){
                            ies.fmSetPortAttribute(logicalPort, FM_PORT_DEF_SWPRI, getYAMLUnsignedInteger(attr, 0));
                        }else if(key == "routable"){
                            ies.fmSetPortAttribute(logicalPort, FM_PORT_ROUTABLE, getYAMLBool(attr, isSpecialPort));
                        }else if(key == "autonegotiation" && isEplPort){
                            ies.fmSetPortAttribute(logicalPort, FM_PORT_AUTONEG, getMapState(getYAMLString(attr, "none"), AUTONEG_MAP));
                        }else if(key == "eth_mode" && isEplPort){
                            ies.fmSetPortAttribute(logicalPort, FM_PORT_ETHERNET_INTERFACE_MODE, getMapState(getYAMLString(attr, "none"), ETHMODE_MAP));
                        }else if (key == "tagging" && attr.is_map()){
                            for(auto attr2 : attr){
                                std::string key2;
                                c4::from_chars(attr2.key(), &key2);

                                if(key2 == "mode"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_TAGGING_MODE, getMapState(getYAMLString(attr2), TAGGING_MODE_MAP));
                                }else if(key2 == "vid"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_DEF_VLAN, getYAMLUnsignedInteger(attr2, 1));
                                }else if(key2 == "priority"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_DEF_PRI, getYAMLUnsignedInteger(attr2, 0));
                                }else if(key2 == "vid2"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_DEF_VLAN2, getYAMLUnsignedInteger(attr2, 1));
                                }else if(key2 == "priority2"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_DEF_PRI2, getYAMLUnsignedInteger(attr2, 0));
                                }else if(key2 == "replace"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_REPLACE_VLAN_FIELDS, getYAMLBool(attr2, false));
                                }else if(key2 == "cfi"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_DEF_CFI, getYAMLUnsignedInteger(attr2, 0));
                                }else if(key2 == "dscp"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_DEF_DSCP, getYAMLUnsignedInteger(attr2, 0));
                                }else if(key2 == "isl_format"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_ISL_TAG_FORMAT, getMapState(getYAMLString(attr2, "none"), ISL_FORMAT_MAP));
                                }

                            }
                        }else if (key == "drop" && attr.is_map()){
                            for(auto attr2 : attr){
                                std::string key2;
                                c4::from_chars(attr2.key(), &key2);

                                if(key2 == "boundary_violation"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_DROP_BV, getYAMLBool(attr2, false));
                                }else if(key2 == "untagged"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_DROP_UNTAGGED, getYAMLBool(attr2, false));
                                }else if(key2 == "tagged"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_DROP_TAGGED, getYAMLBool(attr2, false));
                                }
                            }
                        }else if (key == "loopback" && attr.is_map()){
                            for(auto attr2 : attr){
                                std::string key2;
                                c4::from_chars(attr2.key(), &key2);

                                if(key2 == "suppression"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_LOOPBACK_SUPPRESSION, getYAMLBool(attr2, true));
                                }else if(key2 == "mode" && isEplPort){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_LOOPBACK, getMapState(getYAMLString(attr2, "off"), LOOPBACK_MODE_MAP));
                                }else if(key2 == "fabric"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_FABRIC_LOOPBACK, getMapState(getYAMLString(attr2, "off"), LOOPBACK_MODE_MAP));
                                }
                            }
                        }else if (key == "parser" && attr.is_map()){
                            for(auto attr2 : attr){
                                std::string key2;
                                c4::from_chars(attr2.key(), &key2);

                                if(key2 == "first_vlan2"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_PARSER_VLAN2_FIRST, getYAMLBool(attr2, false));
                                }else if(key2 == "first_vid2"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_MODIFY_VID2_FIRST, getYAMLBool(attr2, false));
                                }else if(key2 == "mode"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_PARSER, getMapState(getYAMLString(attr2, "l4"), PARSER_MODE_MAP));
                                }
                            }
                        }else if (key == "pause" && attr.is_map()){
                            for(auto attr2 : attr){
                                std::string key2;
                                c4::from_chars(attr2.key(), &key2);

                                if(key2 == "rx"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_RX_PAUSE, getYAMLBool(attr2, false));
                                }else if(key2 == "tx"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_TX_PAUSE, getYAMLBool(attr2, false));
                                }
                            }
                        }else if (key == "update" && attr.is_map()){
                            for(auto attr2 : attr){
                                std::string key2;
                                c4::from_chars(attr2.key(), &key2);

                                if(key2 == "ttl"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_UPDATE_TTL, getYAMLBool(attr2, true));
                                }else if(key2 == "dscp"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_UPDATE_DSCP, getYAMLBool(attr2, true));
                                }else if(key2 == "vlan_priority"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_TXVPRI, getYAMLBool(attr2, false));
                                }else if(key2 == "vlan_priority2"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_TXVPRI2, getYAMLBool(attr2, false));
                                }else if(key2 == "routed" && attr2.is_seq()){
                                    fm_uint32 f = 0;
                                    for(auto s : attr2){
                                        if(s == "DMAC"){
                                            f |= FM_PORT_ROUTED_FRAME_UPDATE_DMAC;
                                        }else if(s == "SMAC"){
                                            f |= FM_PORT_ROUTED_FRAME_UPDATE_SMAC;
                                        }else if(s == "VLAN"){
                                            f |= FM_PORT_ROUTED_FRAME_UPDATE_VLAN;
                                        }
                                    }
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_ROUTED_FRAME_UPDATE_FIELDS, f);
                                }
                            }
                        }else if (key == "broadcast" && attr.is_map()){
                            for(auto attr2 : attr){
                                std::string key2;
                                c4::from_chars(attr2.key(), &key2);

                                if(key2 == "pruning"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_BCAST_PRUNING, getYAMLBool(attr2, false));
                                }else if(key2 == "flooding"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_BCAST_FLOODING, getMapState(getYAMLString(attr2, "forward_without_cpu"), BFLOOD_MODE_MAP));
                                }
                            }
                        }else if (key == "multicast" && attr.is_map()){
                            for(auto attr2 : attr){
                                std::string key2;
                                c4::from_chars(attr2.key(), &key2);

                                if(key2 == "pruning"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_MCAST_PRUNING, getYAMLBool(attr2, false));
                                }else if(key2 == "flooding"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_MCAST_FLOODING, getMapState(getYAMLString(attr2, "forward_without_cpu"), MFLOOD_MODE_MAP));
                                }
                            }
                        }else if (key == "unicast" && attr.is_map()){
                            for(auto attr2 : attr){
                                std::string key2;
                                c4::from_chars(attr2.key(), &key2);

                                if(key2 == "pruning"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_UCAST_PRUNING, getYAMLBool(attr2, false));
                                }else if(key2 == "flooding"){
                                    ies.fmSetPortAttribute(logicalPort, FM_PORT_UCAST_FLOODING, getMapState(getYAMLString(attr2, "forward_without_cpu"), UFLOOD_MODE_MAP));
                                }
                            }
                        }else if(key == "mask"){
                            auto switchInfo = ies.fmGetSwitchInfo();
                            fm_bitArray bitArray;
                            if(fmCreateBitArray(&bitArray, switchInfo.numCardPorts) != FM_OK){
                                std::abort();
                            }
                            if(attr.is_seq()){
                                for(auto b : attr){
                                    fmSetBitArrayBit(&bitArray, getYAMLInteger(b), 1);
                                }
                            }else if(attr.val() == "-1"){
                                for(auto i = 0; i < switchInfo.numCardPorts; ++i){
                                    if(i != logicalPort){
                                        fmSetBitArrayBit(&bitArray, i, 1);
                                    }
                                }
                            }

                            if(::fmSetPortAttribute(ies.fmGetSwitch(), logicalPort, FM_PORT_MASK_WIDE, &bitArray) != FM_OK){
                                std::abort();
                            }

                            fmDeleteBitArray(&bitArray);
                        }
                    }
                }
            };

            ies.iteratePorts([&ies, &ports, &applyPortSettings](int logicalPort, int physicalPort, int cpi, bool isInternalPort, bool isPCIePort, bool isSpecialPort, bool isEplPort, bool isCpuPort){
                std::cout << "port#" << logicalPort << " phys#" << physicalPort << " cpi#" << cpi
                << (isInternalPort ? " internal" : "")
                << (isPCIePort ? " pcie" : "")
                << (isSpecialPort ? " special" : "")
                << (isEplPort ? " epl" : "")
                << (isCpuPort ? " cpu" : "")
                << std::endl;


                ies.fmSetPortState(logicalPort, FM_PORT_STATE_UP, 0);
                wait_ms(1000);

                for(const auto& entryKind : std::vector<std::string>({"default", "eth", "pcie", "special", "internal", "cpu"})){
                    if(entryKind != "internal" && isInternalPort){
                        //Specifically disallow setting internal in "default"
                        continue;
                    }else if(entryKind == "eth" && !isEplPort){
                        continue;
                    }else if(entryKind == "pcie" && !isPCIePort){
                        continue;
                    }else if(entryKind == "special" && !isSpecialPort){
                        continue;
                    }else if(entryKind == "cpu" && !isCpuPort){
                        continue;
                    }
                    std::cout << "Applying " << entryKind << " settings" << std::endl;
                    for(auto portEntry : ports){
                        if(portEntry["port"].is_seq()){
                            for(auto e : portEntry["port"]){
                                if(e.val() == entryKind){
                                    applyPortSettings(portEntry, logicalPort, physicalPort, cpi, isInternalPort, isPCIePort, isSpecialPort, isEplPort, isCpuPort);
                                }
                            }
                        }else if(portEntry["port"].val() == entryKind){
                            applyPortSettings(portEntry, logicalPort, physicalPort, cpi, isInternalPort, isPCIePort, isSpecialPort, isEplPort, isCpuPort);
                        }
                    }
                }
                for(auto portEntry : ports){
                    if(portEntry["port"].is_seq()){
                        for(auto e : portEntry["port"]){
                            if(e.val().is_integer() && getYAMLUnsignedInteger(e) == logicalPort){
                                applyPortSettings(portEntry, logicalPort, physicalPort, cpi, isInternalPort, isPCIePort, isSpecialPort, isEplPort, isCpuPort);
                            }
                        }
                    }else if(portEntry["port"].val().is_integer() && getYAMLUnsignedInteger(portEntry["port"]) == logicalPort){
                        std::cout << "Applying port# " << logicalPort << " settings" << std::endl;
                        applyPortSettings(portEntry, logicalPort, physicalPort, cpi, isInternalPort, isPCIePort, isSpecialPort, isEplPort, isCpuPort);
                    }
                }
            });
        }

        if(!root["switch"].is_seed() && root["switch"].is_map()){

            if(!root["switch"]["attributes"].is_seed() && root["switch"]["attributes"].is_map()){
                for(auto attr : root["switch"]["attributes"]){
                    std::string key;
                    c4::from_chars(attr.key(), &key);

                    if(key == "broadcast_flooding"){
                        ies.fmSetSwitchAttribute(FM_BCAST_FLOODING, getMapState(getYAMLString(attr, "forward_without_cpu"), BFLOOD_MODE_MAP));
                    }else if(key == "multicast_flooding"){
                        ies.fmSetSwitchAttribute(FM_MCAST_FLOODING, getMapState(getYAMLString(attr, "forward_without_cpu"), MFLOOD_MODE_MAP));
                    }else if(key == "unicast_flooding"){
                        ies.fmSetSwitchAttribute(FM_UCAST_FLOODING, getMapState(getYAMLString(attr, "forward_without_cpu"), UFLOOD_MODE_MAP));
                    }else if(key == "cpu_port"){
                        ies.fmSetSwitchAttribute(FM_REDIRECT_CPU_TRAFFIC, getYAMLInteger(attr));
                    }else if(key == "frame_aging_milliseconds"){
                        ies.fmSetSwitchAttribute(FM_FRAME_AGING_TIME_MSEC, getYAMLInteger(attr));
                    }else if (key == "vlan_learning" && attr.is_map()){
                        for(auto attr2 : attr){
                            std::string key2;
                            c4::from_chars(attr2.key(), &key2);

                            if(key2 == "vlan" && ies.fmGetSwitchAttribute<fm_int>(FM_VLAN_LEARNING_SHARED_VLAN) == FM_VLAN_LEARNING_MODE_SHARED){
                                ies.fmSetSwitchAttribute(FM_VLAN_LEARNING_SHARED_VLAN, getYAMLUnsignedInteger(attr2, 1));
                            }else if(key2 == "mode"){
                                ies.fmSetSwitchAttribute(FM_VLAN_LEARNING_MODE, getMapState(getYAMLString(attr2, "independent"), VLAN_LEARNING_MODE));
                            }
                        }
                    }else if (key == "lag" && attr.is_map()){
                        for(auto attr2 : attr){
                            std::string key2;
                            c4::from_chars(attr2.key(), &key2);

                            if(key2 == "mode"){
                                ies.fmSetSwitchAttribute(FM_LAG_MODE, getMapState(getYAMLString(attr2, "static"), LAG_MODE));
                            }else if(key2 == "pruning"){
                                ies.fmSetSwitchAttribute(FM_LAG_PRUNING, getYAMLBool(attr2, true));
                            }
                        }
                    }else if (key == "trap" && attr.is_map()){
                        for(auto attr2 : attr){
                            std::string key2;
                            c4::from_chars(attr2.key(), &key2);

                            if(key2 == "cpu_mac"){
                                ies.fmSetSwitchAttribute(FM_CPU_MAC, static_cast<fm_macaddr>(getYAMLUnsignedInteger64(attr2, 0x000000000000)));
                            }else if(key2 == "8021x"){
                                ies.fmSetSwitchAttribute(FM_TRAP_IEEE_8021X, getYAMLBool(attr2, true));
                            }else if(key2 == "bpdu"){
                                ies.fmSetSwitchAttribute(FM_TRAP_IEEE_BPDU, getYAMLBool(attr2, true));
                            }else if(key2 == "lacp"){
                                ies.fmSetSwitchAttribute(FM_TRAP_IEEE_LACP, getYAMLBool(attr2, true));
                            }else if(key2 == "garp"){
                                ies.fmSetSwitchAttribute(FM_TRAP_IEEE_GARP, getYAMLBool(attr2, true));
                            }else if(key2 == "mtu_violation"){
                                ies.fmSetSwitchAttribute(FM_TRAP_MTU_VIOLATIONS, getYAMLBool(attr2, true));
                            }
                        }
                    }else if(key == "stp_mode"){
                        ies.fmSetSwitchAttribute(FM_SPANNING_TREE_MODE, getMapState(getYAMLString(attr, "shared"), STP_MODE));
                    }else if(key == "qos"){
                        for(auto attr2 : attr){
                            std::string key2;
                            c4::from_chars(attr2.key(), &key2);

                            if(key2 == "auto_pause"){
                                auto val = getYAMLBool(attr2, true);
                                if(fmSetSwitchQOS(ies.fmGetSwitch(), FM_AUTO_PAUSE_MODE, 0, &val) != FM_OK){
                                    std::abort();
                                }
                            }
                        }
                    }
                }
            }

            std::cout << std::endl;

        }

        if(!root["acls"].is_seed() && root["acls"].is_map()){
            for(auto aclEntry : root["acls"]){
                auto acl = getYAMLInteger(aclEntry["acl"], 1);

                fm_uint32 scenarios = FM_ACL_SCENARIO_ANY_FRAME_TYPE | FM_ACL_SCENARIO_ANY_ROUTING_TYPE;
                if(!aclEntry["scenarios"].is_seed() && aclEntry["scenarios"].is_seq()){
                    for(auto s : aclEntry["scenarios"]){
                        if(s == "FRAME_NONIP"){
                            scenarios |= FM_ACL_SCENARIO_NONIP;
                        }else if(s == "FRAME_IPV4"){
                            scenarios |= FM_ACL_SCENARIO_IPv4;
                        }else if(s == "FRAME_IPV6"){
                            scenarios |= FM_ACL_SCENARIO_IPv6;
                        }else if(s == "FRAME_ANY"){
                            scenarios |= FM_ACL_SCENARIO_ANY_FRAME_TYPE;
                        }else if(s == "ROUTING_SWITCHED"){
                            scenarios |= FM_ACL_SCENARIO_SWITCHED;
                        }else if(s == "ROUTING_UNICAST"){
                            scenarios |= FM_ACL_SCENARIO_UNICAST_ROUTED;
                        }else if(s == "ROUTING_MULTICAST"){
                            scenarios |= FM_ACL_SCENARIO_MULTICAST_ROUTED;
                        }else if(s == "ROUTING_ANY"){
                            scenarios |= FM_ACL_SCENARIO_ANY_ROUTING_TYPE;
                        }else if(s == "GLORT_SWITCHED"){
                            scenarios |= FM_ACL_SCENARIO_SWITCHED_GLORT;
                        }else if(s == "GLORT_UNICAST"){
                            scenarios |= FM_ACL_SCENARIO_UCAST_ROUTED_GLORT;
                        }else if(s == "GLORT_MULTICAST"){
                            scenarios |= FM_ACL_SCENARIO_MCAST_ROUTED_GLORT;
                        }else if(s == "GLORT_ANY"){
                            scenarios |= FM_ACL_SCENARIO_ANY_ROUTING_GLORT_TYPE;
                        }else if(s == "SPECIAL_GLORT_FTYPE"){
                            scenarios |= FM_ACL_SCENARIO_SPECIAL_GLORT;
                        }
                    }
                }

                auto precedence = getYAMLInteger(aclEntry["precedence"], FM_ACL_DEFAULT_PRECEDENCE);

                if(fmCreateACLExt(ies.fmGetSwitch(), acl, scenarios, precedence) != FM_OK){
                    std::abort();
                }

                if (!aclEntry["rules"].is_seed() && aclEntry["rules"].is_seq()){
                    for(auto ruleEntry : aclEntry["rules"]){
                        auto rule = getYAMLInteger(ruleEntry["rule"]);
                        fm_aclCondition condition{};
                        fm_aclValue value{};
                        fm_aclActionExt action{};
                        fm_aclParamExt param{};


#define ACL_VMASK(c, k)                                 \
condition |= c;                                         \
setACLMatchValueMask(attr2, value. k, value. k ## Mask);

                        for(auto attr : ruleEntry["match"]) {
                            std::string key;
                            c4::from_chars(attr.key(), &key);

                            if (key == "logical" && attr.is_map()){
                                for(auto attr2 : attr){
                                    std::string key2;
                                    c4::from_chars(attr2.key(), &key2);

                                    if(key2 == "switch_priority"){
                                        ACL_VMASK(FM_ACL_MATCH_SWITCH_PRIORITY, switchPri);
                                    }else if(key2 == "scenario_flags"){
                                        ACL_VMASK(FM_ACL_MATCH_SCENARIO_FLAGS, scenarioFlags);
                                    }else if(key2 == "src_port"){
                                        condition |= FM_ACL_MATCH_SRC_PORT;
                                        typeof(value.logicalPort) mask;//[sizeof(value.fType) / sizeof(std::remove_extent<typeof(value.fType)>)];
                                        setACLMatchValueMask(attr2, value.logicalPort, mask);
                                    }
                                }
                            }else if (key == "l2" && attr.is_map()){
                                for(auto attr2 : attr){
                                    std::string key2;
                                    c4::from_chars(attr2.key(), &key2);

                                    if(key2 == "src_mac"){
                                        ACL_VMASK(FM_ACL_MATCH_SRC_MAC, src);
                                    }else if(key2 == "dst_mac"){
                                        ACL_VMASK(FM_ACL_MATCH_DST_MAC, dst);
                                    }else if(key2 == "ethertype"){
                                        ACL_VMASK(FM_ACL_MATCH_ETHERTYPE, ethType);
                                    }else if(key2 == "vlan"){
                                        ACL_VMASK(FM_ACL_MATCH_VLAN, vlanId);
                                    }else if(key2 == "priority"){
                                        ACL_VMASK(FM_ACL_MATCH_PRIORITY, vlanPri);
                                    }else if(key2 == "vlan2"){
                                        ACL_VMASK(FM_ACL_MATCH_VLAN2, vlanId2);
                                    }else if(key2 == "priority2"){
                                        ACL_VMASK(FM_ACL_MATCH_PRIORITY2, vlanPri2);
                                    }else if(key2 == "isl_ftype"){
                                        condition |= FM_ACL_MATCH_ISL_FTYPE;
                                        typeof(value.fType) mask;//[sizeof(value.fType) / sizeof(std::remove_extent<typeof(value.fType)>)];
                                        setACLMatchValueMask(attr2, value.fType, mask);
                                    }else if(key2 == "isl_user"){
                                        ACL_VMASK(FM_ACL_MATCH_ISL_USER, islUser);
                                    }
                                }
                            }else if (key == "l3" && attr.is_map()){
                                for(auto attr2 : attr){
                                    std::string key2;
                                    c4::from_chars(attr2.key(), &key2);

                                    if(key2 == "src_ip"){
                                        ACL_VMASK(FM_ACL_MATCH_SRC_IP, srcIp);
                                    }else if(key2 == "dst_ip"){
                                        ACL_VMASK(FM_ACL_MATCH_DST_IP, dstIp);
                                    }else if(key2 == "ttl"){
                                        ACL_VMASK(FM_ACL_MATCH_TTL, ttl);
                                    }else if(key2 == "protocol"){
                                        ACL_VMASK(FM_ACL_MATCH_PROTOCOL, protocol);
                                    }else if(key2 == "flags"){
                                        ACL_VMASK(FM_ACL_MATCH_FLAGS, flags);
                                    }else if(key2 == "dscp"){
                                        ACL_VMASK(FM_ACL_MATCH_DSCP, dscp);
                                    }else if(key2 == "non_ip_payload"){
                                        ACL_VMASK(FM_ACL_MATCH_NON_IP_PAYLOAD, nonIPPayload);
                                    }else if(key2 == "dpi"){
                                        ACL_VMASK(FM_ACL_MATCH_L4_DEEP_INSPECTION_EXT, L4DeepInspectionExt);
                                    }else if(key2 == "frame_type"){
                                        condition |= FM_ACL_MATCH_FRAME_TYPE;
                                        value.frameType = (fm_aclFrameType) getMapState(getYAMLString(attr2), ACL_FRAME_TYPE_MAP);
                                    }
                                }
                            }else if (key == "l4" && attr.is_map()){
                                for(auto attr2 : attr){
                                    std::string key2;
                                    c4::from_chars(attr2.key(), &key2);

                                    if(key2 == "src_port"){
                                        condition |= FM_ACL_MATCH_L4_SRC_PORT_WITH_MASK;
                                        setACLMatchValueMask(attr2, value.L4SrcStart, value.L4SrcMask);
                                    }else if(key2 == "dst_port"){
                                        condition |= FM_ACL_MATCH_L4_DST_PORT_WITH_MASK;
                                        setACLMatchValueMask(attr2, value.L4DstStart, value.L4DstMask);
                                    }else if(key2 == "tcp_flags"){
                                        ACL_VMASK(FM_ACL_MATCH_TCP_FLAGS, tcpFlags);
                                    }else if(key2 == "non_ip_payload"){
                                        ACL_VMASK(FM_ACL_MATCH_NON_IP_PAYLOAD, nonIPPayload);
                                    }
                                }
                            }
                        }


                        if(!ruleEntry["action"].is_seed() && ruleEntry["action"].is_seq()){
                            for(auto v : ruleEntry["action"]) {
                                if(v == "permit" || v == "allow"){
                                    action |= FM_ACL_ACTIONEXT_PERMIT;
                                }else if(v == "deny"){
                                    action |= FM_ACL_ACTIONEXT_DENY;
                                }else if(v == "trap"){
                                    action |= FM_ACL_ACTIONEXT_TRAP;
                                }else if(v == "log"){
                                    action |= FM_ACL_ACTIONEXT_LOG;
                                }else if(v == "count"){
                                    action |= FM_ACL_ACTIONEXT_COUNT;
                                }else if(v == "trap_always"){
                                    action |= FM_ACL_ACTIONEXT_TRAP_ALWAYS;
                                }else if(v == "noroute"){
                                    action |= FM_ACL_ACTIONEXT_NOROUTE;
                                }else if(v == "pop_vlan"){
                                    action |= FM_ACL_ACTIONEXT_POP_VLAN;
                                }else if(v == "capture_egress_timestamp"){
                                    action |= FM_ACL_ACTIONEXT_CAPTURE_EGRESS_TIMESTAMP;
                                }
                            }
                        }


                        if(ruleEntry["set"].is_seed() && ruleEntry["set"].is_map()){
                            for(auto attr : ruleEntry["action"]) {
                                std::string key;
                                c4::from_chars(attr.key(), &key);

                                if(key == "police"){
                                    action |= FM_ACL_ACTIONEXT_POLICE;
                                    param.policer = getYAMLInteger(attr);
                                }else if(key == "vlan"){
                                    action |= FM_ACL_ACTIONEXT_SET_VLAN;
                                    param.vlan = getYAMLInteger(attr);
                                }else if(key == "priority"){
                                    action |= FM_ACL_ACTIONEXT_SET_VLAN_PRIORITY;
                                    param.vlanPriority = getYAMLInteger(attr);
                                }else if(key == "switch_priority"){
                                    action |= FM_ACL_ACTIONEXT_SET_SWITCH_PRIORITY;
                                    param.switchPriority = getYAMLInteger(attr);
                                }else if(key == "dscp"){
                                    action |= FM_ACL_ACTIONEXT_SET_DSCP;
                                    param.dscp = getYAMLInteger(attr);
                                }else if(key == "user"){
                                    action |= FM_ACL_ACTIONEXT_SET_USER;
                                    param.user = getYAMLInteger(attr);
                                }else if(key == "user_mask"){
                                    action |= FM_ACL_ACTIONEXT_SET_USER;
                                    param.userMask = getYAMLInteger(attr);
                                }else if(key == "load_balance"){
                                    action |= FM_ACL_ACTIONEXT_LOAD_BALANCE;
                                    param.lbgNumber = getYAMLInteger(attr);
                                }else if(key == "redirect"){
                                    action |= FM_ACL_ACTIONEXT_REDIRECT;
                                    param.logicalPort = getYAMLInteger(attr);
                                }else if(key == "flood_destination"){
                                    action |= FM_ACL_ACTIONEXT_SET_FLOOD_DEST;
                                    param.logicalPort = getYAMLInteger(attr);
                                }else if(key == "route_ecmp"){
                                    action |= FM_ACL_ACTIONEXT_ROUTE;
                                    param.groupId = getYAMLInteger(attr);
                                }else if(key == "mirror_group"){
                                    action |= FM_ACL_ACTIONEXT_MIRROR_GRP;
                                    param.mirrorGrp = getYAMLInteger(attr);
                                }else if(key == "precedence"){
                                    action |= FM_ACL_ACTIONEXT_SET_PRECEDENCE;
                                    param.precedence = getYAMLInteger(attr);
                                }else if(key == "push_vlan"){
                                    action |= FM_ACL_ACTIONEXT_PUSH_VLAN;
                                    param.vlan = getYAMLInteger(attr);
                                }else if(key == "update_or_add_vlan"){
                                    action |= FM_ACL_ACTIONEXT_UPD_OR_ADD_VLAN;
                                    param.vlan = getYAMLInteger(attr);
                                }
                            }
                        }



                        if(fmAddACLRuleExt(ies.fmGetSwitch(), acl, rule, condition, &value, action, &param) != FM_OK){
                            std::abort();
                        }
                    }
#undef ACL_VMASK
                }
            }
        }

        if(!root["router"].is_seed() && root["router"].is_map()){

            if(!root["router"]["attributes"].is_seed() && root["router"]["attributes"].is_map()){

                for(auto attr : root["router"]["attributes"]) {
                    std::string key;
                    c4::from_chars(attr.key(), &key);

                    if (key == "mac" && attr.is_map()){
                        for(auto attr2 : attr){
                            std::string key2;
                            c4::from_chars(attr2.key(), &key2);

                            if(key2 == "real"){
                                ies.fmSetRouterAttribute(FM_ROUTER_PHYSICAL_MAC_ADDRESS, static_cast<fm_macaddr>(getYAMLUnsignedInteger64(attr2, 0x000000000000)));
                            }else if(key2 == "virtual1"){
                                ies.fmSetRouterAttribute(FM_ROUTER_VIRTUAL_MAC_ADDRESS, static_cast<fm_macaddr>(getYAMLUnsignedInteger64(attr2, 0x000000000000)));
                            }else if(key2 == "virtual2"){
                                ies.fmSetRouterAttribute(FM_ROUTER_VIRTUAL_MAC_ADDRESS_2, static_cast<fm_macaddr>(getYAMLUnsignedInteger64(attr2, 0x000000000000)));
                            }
                        }
                    }
                }
            }

            if(!root["router"]["routers"].is_seed() && root["router"]["routers"].is_seq()){

                for(auto routerEntry : root["switch"]["routers"]) {
                    auto vrid = getYAMLInteger(routerEntry["kind"], 0);
                    if(vrid != 0){
                        if(fmCreateVirtualRouter(ies.fmGetSwitch(), vrid) != FM_OK){
                            std::abort();
                        }
                    }
                    if(!routerEntry["mac_mode"].is_seed()){
                        if(fmSetRouterMacMode(ies.fmGetSwitch(), vrid, (fm_routerMacMode) getMapState(getYAMLString(routerEntry["mac_mode"]), ROUTER_MAC_MODE_MAP)) != FM_OK){
                            std::abort();
                        }
                    }

                    if(!routerEntry["routes"].is_seed() && routerEntry["routes"].is_seq()){
                        for(auto routeEntry : routerEntry["routes"]) {
                            fm_routeEntry entry;
                            entry.routeType = (fm_routeType) getMapState(getYAMLString(routeEntry["type"], "unicast"), ROUTE_ENTRY_TYPE_MAP);
                            entry.data.unicast.dstAddr = stringToIpAddr(getYAMLString(routeEntry["destination"]).c_str());
                            entry.data.unicast.prefixLength = getYAMLInteger(routeEntry["mask"], entry.data.unicast.dstAddr.isIPv6 ? 64 : 32);
                            entry.data.unicast.nextHop = stringToIpAddr(getYAMLString(routeEntry["next_hop"]).c_str());
                            entry.data.unicast.interfaceAddr = stringToIpAddr(getYAMLString(routeEntry["interface"]).c_str());
                            entry.data.unicast.vlan = getYAMLInteger(routeEntry["vlan"], 0);
                            entry.data.unicast.vrid = vrid;
                            if(fmAddRoute(ies.fmGetSwitch(), &entry, FM_ROUTE_STATE_UP) != FM_OK){
                                std::abort();
                            }
                        }
                    }
                }
            }
        }

        std::cout << "Switch is UP, all ports are now enabled" << std::endl;

        return true;
    }

    return false;
}
